#include "LLL.h"
#include <time.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>


int VERBOSE=0;
int EARLY_PREC_DETECT=0;
int DEEP_LOVASZ=0;

volatile int userwantexit=0;

typedef struct _lllparam {
  LatEntry *Lattice;
  LatEntry *ScalProd;
  LatEntry *ExactScals;
  LatEntry *ExactScalsNoise;
  GramEntry *GramR;
  GramEntry *GramM;
  GramEntry *GramS;
  GramEntry LLLCste;
  int spacedim;
  int allocdim;
  int dim;
  int *muinvalidfrom;
  GramEntry mu;
  GramEntry Tmp;
  GramEntry Quo;
  GramEntry Scal;
  GramEntry mubis;
  GramEntry muter;
  GramEntry Scalneg;
  GramEntry Scalpos;
  GramEntry MinusOne;
  LatEntry *LatTmp_ptr;
  LatEntry *exScalar_ptr;
  LatEntry *RoundedQuo_ptr;
} LLL_PARAMS;


void AddLatEntryToGramNoise(GramEntry Gr, LatEntry Er, GramEntry noise, GramEntry tmp)
{
  long nbbitsmul, offset, limb_offset, errorshift;
  mp_limb_t errordrop;
  
  if ((Er->_mp_size)<0) {
    fprintf(stderr,"Negative Error to add ??? \n");
    exit(1);
  }

  flti_zero(noise);

  if ((Er->_mp_size)==0) return;
  nbbitsmul=my_mpn_bitnum(Er->_mp_d,  Er->_mp_size);

  if (nbbitsmul<=MAX_ERROR_PREC) {
    noise->errormant=(Er->_mp_d)[0];
    errorshift=0;
    errordrop=0;
  }
  else {
    errorshift=nbbitsmul-MAX_ERROR_PREC;
    limb_offset=nbbitsmul/GMP_LIMB_BITS;
    offset=nbbitsmul%GMP_LIMB_BITS;
    if (offset>=MAX_ERROR_PREC) {
      noise->errormant=(Er->_mp_d)[limb_offset]>>(offset-MAX_ERROR_PREC);
      errordrop=1;
    }
    else {
      noise->errormant=((Er->_mp_d)[limb_offset]<<(MAX_ERROR_PREC-offset)) |
	((Er->_mp_d)[limb_offset-1]>>(GMP_LIMB_BITS+offset-MAX_ERROR_PREC));
      errordrop=1;
    }
    noise->exponent=errorshift;
  }


  if (errordrop) {
    noise->errormant++;
  }

  if ((noise->errormant)>=(1UL<<MAX_ERROR_PREC)) {
    errordrop=(noise->errormant)&1;
    (noise->errormant)>>=1;
    if (errordrop) (noise->errormant)++;    
    (noise->exponent)++;
  }

  flti_add(tmp,Gr,noise);
  flti_cpy(Gr,tmp);
}


int LowLevelExactMinusDotProduct(GramEntry Scal, LatEntry BaseVal,
				 GramEntry *Vec1, GramEntry *Vec2, int len)
{
  mp_limb_t mantiss[2*MAX_LIMBS_PREC];
  mp_limb_t errormanti[MAX_LIMBS_PREC+3];
  mp_limb_t mantiss_pos_accum[2*MAX_LIMBS_PREC+1];
  mp_limb_t mantiss_neg_accum[2*MAX_LIMBS_PREC+1];
  mp_limb_t erroraccum[MAX_LIMBS_PREC+2];
  mp_limb_t drop, errordrop, carry;

  long glob_exponent, offset, limb_offset, nbbitsmul;
  long mantissashift,errorshift;
  int i,sgn;
  
  long mpzsize, full_mul_size, full_error_size, maxexponent;

  full_mul_size=2*flti_glob_nblimbs+1;
  full_error_size=flti_glob_nblimbs+1;
  mpn_zero (erroraccum, full_error_size);

  maxexponent=-full_mul_size*GMP_LIMB_BITS;
  for(i=0;i<len;i++) {
    if (((Vec1[i]->sign)||(Vec1[i]->errormant)) &&
	((Vec2[i]->sign)||(Vec2[i]->errormant))) {
      if ((Vec1[i]->exponent)+(Vec2[i]->exponent)>maxexponent) {
	maxexponent=(Vec1[i]->exponent)+(Vec2[i]->exponent);
      }
    }
  }

  mpzsize=BaseVal->_mp_size;
  if (mpzsize==0) {
    glob_exponent=maxexponent;
    mpn_zero (mantiss_pos_accum, full_mul_size);
    mpn_zero (mantiss_neg_accum, full_mul_size);
  }
  else {
    if (mpzsize>0) {
      glob_exponent=(mpzsize-full_mul_size)*GMP_LIMB_BITS;
      if (mpzsize > full_mul_size) {
	mpn_copyi(mantiss_pos_accum, (BaseVal->_mp_d)+mpzsize-full_mul_size, full_mul_size);
	mpn_zero (mantiss_neg_accum, full_mul_size);
	erroraccum[0]=1;
      }
      else {
	mpn_zero (mantiss_pos_accum, full_mul_size-mpzsize);
	mpn_copyi(mantiss_pos_accum+full_mul_size-mpzsize, (BaseVal->_mp_d), mpzsize);
	mpn_zero (mantiss_neg_accum, full_mul_size);
      }
    }
    else {
      mpzsize=-mpzsize;
      glob_exponent=(mpzsize-full_mul_size)*GMP_LIMB_BITS;
      if (mpzsize > full_mul_size) {
	mpn_copyi(mantiss_neg_accum, (BaseVal->_mp_d)+mpzsize-full_mul_size, full_mul_size);
	mpn_zero (mantiss_pos_accum, full_mul_size);
	erroraccum[0]=1;
      }
      else {
	mpn_zero (mantiss_neg_accum, full_mul_size-mpzsize);
	mpn_copyi(mantiss_neg_accum+full_mul_size-mpzsize, (BaseVal->_mp_d), mpzsize);
	mpn_zero (mantiss_pos_accum, full_mul_size);
      }
    }

    if (maxexponent>glob_exponent) {
      // fprintf(stderr,"maxexponent=%d vs glob_exponent=%d\n", maxexponent,glob_exponent);
      offset=maxexponent-glob_exponent;
      limb_offset=offset/GMP_LIMB_BITS;
      offset=offset%GMP_LIMB_BITS;
      if (limb_offset<full_mul_size) {
	if (offset) {
	  mpn_rshift(mantiss_pos_accum, mantiss_pos_accum+limb_offset,
		     full_mul_size-limb_offset, offset);
	  mpn_rshift(mantiss_neg_accum, mantiss_neg_accum+limb_offset,
		     full_mul_size-limb_offset, offset);
	}
	else {
	  mpn_copyi(mantiss_pos_accum, mantiss_pos_accum+limb_offset,
		     full_mul_size-limb_offset);
	  mpn_copyi(mantiss_neg_accum, mantiss_neg_accum+limb_offset,
		     full_mul_size-limb_offset);
	}
	mpn_zero (mantiss_pos_accum+full_mul_size-limb_offset, limb_offset);
	mpn_zero (mantiss_neg_accum+full_mul_size-limb_offset, limb_offset);
      }
      else {
	mpn_zero (mantiss_pos_accum, full_mul_size);
	mpn_zero (mantiss_neg_accum, full_mul_size);
      }
      erroraccum[0]=1;
      glob_exponent=maxexponent;
    }
  }
  for(i=0;i<len;i++) {
    mpn_mul_n (mantiss, Vec1[i]->mantissa, Vec2[i]->mantissa, flti_glob_nblimbs);
    mpn_zero(errormanti,flti_glob_nblimbs+1);
    errormanti[0]=(Vec1[i]->errormant)*(Vec2[i]->errormant);
    if ((Vec1[i]->errormant)&&(Vec2[i]->sign)) {
      errormanti[flti_glob_nblimbs]+=mpn_addmul_1 (errormanti, Vec2[i]->mantissa, flti_glob_nblimbs,
						   (mp_limb_t) (Vec1[i]->errormant));
    }
    if ((Vec2[i]->errormant)&&(Vec1[i]->sign)) {
      errormanti[flti_glob_nblimbs]+=mpn_addmul_1 (errormanti, Vec1[i]->mantissa, flti_glob_nblimbs,
						   (mp_limb_t) (Vec2[i]->errormant));
    }

    offset=glob_exponent-(Vec1[i]->exponent)-(Vec2[i]->exponent);
    drop=0;
    errordrop=0;

    limb_offset=offset/GMP_LIMB_BITS;
    offset=offset%GMP_LIMB_BITS;
    if (limb_offset<2*flti_glob_nblimbs) {
      //if (!my_mpn_zero_p(mantiss,  limb_offset))
	drop=1;
      if (offset) {
	if (mpn_rshift(mantiss+limb_offset, mantiss+limb_offset,
		       2*flti_glob_nblimbs-limb_offset, offset)) drop=1;
      }
      if ((Vec1[i]->sign)!=(Vec2[i]->sign)) {
	carry=mpn_add (mantiss_pos_accum, mantiss_pos_accum, full_mul_size,
		       mantiss+limb_offset, 2*flti_glob_nblimbs-limb_offset);
	if (carry) {
	  mantiss_pos_accum[full_mul_size]=carry;
	  mantiss_neg_accum[full_mul_size]=0;
	  full_mul_size++;
	}
      }
      else {
	carry=mpn_add (mantiss_neg_accum, mantiss_neg_accum, full_mul_size,
		       mantiss+limb_offset, 2*flti_glob_nblimbs-limb_offset);
	if (carry) {
	  mantiss_pos_accum[full_mul_size]=0;
	  mantiss_neg_accum[full_mul_size]=carry;
	  full_mul_size++;
	}
      }
    }
    else {
      //if (!my_mpn_zero_p(mantiss,  2*flti_glob_nblimbs))
	drop=1;
    }



    if (limb_offset<flti_glob_nblimbs+1) {
      //if (!my_mpn_zero_p(errormanti,  limb_offset))
	errordrop=1;
      if (offset) {
	if (mpn_rshift(errormanti+limb_offset, errormanti+limb_offset,
		       flti_glob_nblimbs+1-limb_offset, offset)) drop=1;
      }
      carry=mpn_add (erroraccum, erroraccum, full_error_size,
		     errormanti+limb_offset, flti_glob_nblimbs+1-limb_offset);
      if (carry) {
	erroraccum[full_error_size]=carry;
	full_error_size++;
	if (full_error_size>MAX_LIMBS_PREC) {
	  fprintf(stderr,"Error too large ?? \n");
	  exit(1);
	}
      }
    }
    else {
      //if (!my_mpn_zero_p(errormanti,  flti_glob_nblimbs+1))
	errordrop=1;
    }

    if (drop||errordrop) {
      carry=mpn_add_1 (erroraccum, erroraccum, full_error_size, (mp_limb_t) 1UL);
      if (carry) {
	erroraccum[full_error_size]=carry;
	full_error_size++;
      }
    }
  }
  
  
  drop=0;
  errordrop=0;

  sgn=mpn_cmp(mantiss_pos_accum, mantiss_neg_accum, full_mul_size);
  if (sgn>0) {
    Scal->sign=1;
    mpn_sub_n (mantiss_pos_accum, mantiss_pos_accum,  mantiss_neg_accum, full_mul_size);
  }
  else if (sgn<0) {
    Scal->sign=-1;
    mpn_sub_n (mantiss_pos_accum, mantiss_neg_accum,  mantiss_pos_accum, full_mul_size);
  }
  else {
    Scal->sign=0;
    mpn_zero (mantiss_pos_accum, full_mul_size);
  }
  nbbitsmul=my_mpn_bitnum(mantiss_pos_accum, full_mul_size);
  if (nbbitsmul<=flti_glob_prec) {
    mpn_copyi(Scal->mantissa, mantiss_pos_accum, flti_glob_nblimbs);
    mantissashift=0;
  }
  else {
    mantissashift=nbbitsmul-flti_glob_prec;
    limb_offset=mantissashift/GMP_LIMB_BITS;
    offset=mantissashift%GMP_LIMB_BITS;
    if (offset) {
      mpn_lshift (mantiss_pos_accum, mantiss_pos_accum,
		  limb_offset+flti_glob_nblimbs+1, GMP_LIMB_BITS-offset);
      limb_offset++;
    }
    mpn_copyi(Scal->mantissa, mantiss_pos_accum+limb_offset,flti_glob_nblimbs);
    // if (!my_mpn_zero_p(mantiss_pos_accum,limb_offset)) drop=1; else drop=0;
    drop=1;
  }

  nbbitsmul=my_mpn_bitnum(erroraccum,  full_error_size);
  if (nbbitsmul<=MAX_ERROR_PREC) {
    Scal->errormant=erroraccum[0];
    errorshift=0;
    errordrop=0;
  }
  else {
    errorshift=nbbitsmul-MAX_ERROR_PREC;
    limb_offset=errorshift/GMP_LIMB_BITS;
    offset=errorshift%GMP_LIMB_BITS;
    erroraccum[(nbbitsmul/GMP_LIMB_BITS)+1]=0;
    if (offset) {
      limb_offset++;
      mpn_lshift (erroraccum, erroraccum,
		  limb_offset+1, GMP_LIMB_BITS-offset);
    }
    Scal->errormant=erroraccum[limb_offset];
    //    if (!my_mpn_zero_p(erroraccum,limb_offset)) errordrop=1; else errordrop=0;
    errordrop=1;
  }
  
  if (mantissashift>errorshift) {
    Scal->exponent=glob_exponent+mantissashift;
    if ((mantissashift-errorshift)>=MAX_ERROR_PREC) {
      if (Scal->errormant) errordrop=1;
      Scal->errormant=0;
    }
    else {
      offset=mantissashift-errorshift;
      if ((Scal->errormant)&((1UL<<offset)-1)) errordrop=1;
      Scal->errormant>>=offset;
    }
  }
  else if (errorshift>mantissashift) {
    Scal->exponent=glob_exponent+errorshift;
    offset=errorshift-mantissashift;
    if (offset>=flti_glob_prec) {
      //if (!my_mpn_zero_p(Scal->mantissa, flti_glob_nblimbs))
	drop=1;
      mpn_zero(Scal->mantissa, flti_glob_nblimbs);
    }
    else {
      limb_offset=offset/GMP_LIMB_BITS;
      offset=offset%GMP_LIMB_BITS;
      
      //if (!my_mpn_zero_p(Scal->mantissa,  limb_offset))
	drop=1;
      if (offset) {
	if (mpn_rshift(Scal->mantissa, (Scal->mantissa)+limb_offset,
		       flti_glob_nblimbs-limb_offset, offset)) drop=1;
      }
      else {
	mpn_copyi(Scal->mantissa,(Scal->mantissa)+limb_offset, flti_glob_nblimbs-limb_offset);
      }
      mpn_zero((Scal->mantissa)+flti_glob_nblimbs-limb_offset,limb_offset);
    }
  }
  else { // errorshift == mantissashift
    Scal->exponent=glob_exponent+mantissashift;
  }

  if (drop||errordrop) {
    Scal->errormant++;
  }
  
  while ((Scal->errormant)>=(1UL<<MAX_ERROR_PREC)) {
    drop=mpn_rshift (Scal->mantissa, Scal->mantissa, flti_glob_nblimbs, 1);
    errordrop=(Scal->errormant)&1;
    (Scal->errormant)>>=1;
    if (errordrop||drop) (Scal->errormant)++;    
    (Scal->exponent)++;
  }

  if (my_mpn_zero_p(Scal->mantissa, flti_glob_nblimbs)) Scal->sign=0;
  return 0;
}


void FullUpdate(int j,int k, int maxk, LLL_PARAMS Params)
{
  int i;
  LatEntry *Lptr, *Ljptr;
  unsigned long int smallquo;
  unsigned long int smallquosquare;

  LatEntry *Lattice, *ExactScals, *RoundedQuo, *LatTmp;
  GramEntry *GramM, *GramR, Tmp, Scalneg, Quo;
  int spacedim, allocdim;

  Lattice=Params.Lattice; ExactScals=Params.ExactScals; GramR=Params.GramR;
  GramM=Params.GramM; spacedim=Params.spacedim;  Tmp=Params.Tmp;
  allocdim=Params.allocdim; Scalneg=Params.Scalneg;
  LatTmp=Params.LatTmp_ptr; RoundedQuo=Params.RoundedQuo_ptr;
  Quo=Params.Quo;

  smallquo=((*RoundedQuo)->_mp_d)[0];
  if ((((*RoundedQuo)->_mp_size)==1)&&(smallquo<=0xffffffffUL)) {
    for(i=0;i<j;i++) {
      GramMulSmall(Tmp,*(GramM+j*allocdim+i),smallquo);
      GramCpy(Scalneg,*(GramM+k*allocdim+i)); // Use Scalneg as temporary
      GramSub(*(GramM+k*allocdim+i),Scalneg,Tmp);
      GramMulSmall(Tmp,*(GramR+j*allocdim+i),smallquo);
      GramCpy(Scalneg,*(GramR+k*allocdim+i)); // Use Scalneg as temporary
      GramSub(*(GramR+k*allocdim+i),Scalneg,Tmp);
    }
    GramCpy(Tmp,*(GramM+k*allocdim+i));
    GramSub(*(GramM+k*allocdim+i),Tmp,Quo);

    smallquosquare=smallquo*smallquo;
    Lptr=Lattice+k*spacedim;
    Ljptr=Lattice+j*spacedim;
    for(i=0;i<spacedim;i++) {
      if (LatNonZero(*Ljptr)) {
	mpz_mul_ui(*LatTmp,*Ljptr,smallquo);
	LatSub(*Lptr,*Lptr,*LatTmp);
      }
      Lptr++; Ljptr++;
    }
    
    //  Update exact scalar products
    Lptr=ExactScals+k*allocdim+k;
    Ljptr=ExactScals+k*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      mpz_mul_ui(*LatTmp,*Ljptr,smallquo<<1);
      LatSub(*Lptr,*Lptr,*LatTmp);
    }
    Ljptr=ExactScals+j*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      mpz_mul_ui(*LatTmp,*Ljptr,smallquosquare);
      LatAdd(*Lptr,*Lptr,*LatTmp);
    }
    
    for(i=0;i<=maxk;i++) {
      if (i==k) continue;
      if (i<k) 
	Lptr=ExactScals+k*allocdim+i;
      else
	Lptr=ExactScals+i*allocdim+k;
      if (i<j) 
	Ljptr=ExactScals+j*allocdim+i;
      else
	Ljptr=ExactScals+i*allocdim+j;
      if (LatNonZero(*Ljptr)) {
	mpz_mul_ui(*LatTmp,*Ljptr,smallquo);
	LatSub(*Lptr,*Lptr,*LatTmp);
      }
    }
  }
  else if ((((*RoundedQuo)->_mp_size)==-1)&&(smallquo<=0xffffffffUL)) {
    for(i=0;i<j;i++) {
      GramMulSmall(Tmp,*(GramM+j*allocdim+i),smallquo);
      GramCpy(Scalneg,*(GramM+k*allocdim+i)); // Use Scalneg as temporary
      GramAdd(*(GramM+k*allocdim+i),Scalneg,Tmp);
      GramMulSmall(Tmp,*(GramR+j*allocdim+i),smallquo);
      GramCpy(Scalneg,*(GramR+k*allocdim+i)); // Use Scalneg as temporary
      GramAdd(*(GramR+k*allocdim+i),Scalneg,Tmp);
    }
    GramCpy(Tmp,*(GramM+k*allocdim+i));
    GramSub(*(GramM+k*allocdim+i),Tmp,Quo);

    GramMul(Tmp,Quo,*(GramR+j*allocdim+j));
    GramCpy(Scalneg,*(GramR+k*allocdim+j)); // Use Scalneg as temporary
    GramSub(*(GramR+k*allocdim+j),Scalneg,Tmp);
    
    smallquosquare=smallquo*smallquo;
    Lptr=Lattice+k*spacedim;
    Ljptr=Lattice+j*spacedim;
    for(i=0;i<spacedim;i++) {
      if (LatNonZero(*Ljptr)) {
	mpz_mul_ui(*LatTmp,*Ljptr,smallquo);
	LatAdd(*Lptr,*Lptr,*LatTmp);
      }
      Lptr++; Ljptr++;
    }
    
    //  Update exact scalar products
    Lptr=ExactScals+k*allocdim+k;
    Ljptr=ExactScals+k*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      mpz_mul_ui(*LatTmp,*Ljptr,smallquo<<1);
      LatAdd(*Lptr,*Lptr,*LatTmp);
    }
    Ljptr=ExactScals+j*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      mpz_mul_ui(*LatTmp,*Ljptr,smallquosquare);
      LatAdd(*Lptr,*Lptr,*LatTmp);
    }
    
    for(i=0;i<=maxk;i++) {
      if (i==k) continue;
      if (i<k) 
	Lptr=ExactScals+k*allocdim+i;
      else
	Lptr=ExactScals+i*allocdim+k;
      if (i<j) 
	Ljptr=ExactScals+j*allocdim+i;
      else
	Ljptr=ExactScals+i*allocdim+j;
      if (LatNonZero(*Ljptr)) {
	mpz_mul_ui(*LatTmp,*Ljptr,smallquo);
	LatAdd(*Lptr,*Lptr,*LatTmp);
      }
    }
  }
  else {
    for(i=0;i<j;i++) {
      GramMul(Tmp,Quo,*(GramM+j*allocdim+i));
      GramCpy(Scalneg,*(GramM+k*allocdim+i)); // Use Scalneg as temporary
      GramSub(*(GramM+k*allocdim+i),Scalneg,Tmp);
      GramMul(Tmp,Quo,*(GramR+j*allocdim+i));
      GramCpy(Scalneg,*(GramR+k*allocdim+i)); // Use Scalneg as temporary
      GramSub(*(GramR+k*allocdim+i),Scalneg,Tmp);
    }
    GramCpy(Tmp,*(GramM+k*allocdim+i));
    GramSub(*(GramM+k*allocdim+i),Tmp,Quo);

    GramMul(Tmp,Quo,*(GramR+j*allocdim+j));
    GramCpy(Scalneg,*(GramR+k*allocdim+j)); // Use Scalneg as temporary
    GramSub(*(GramR+k*allocdim+j),Scalneg,Tmp);

  
    Lptr=Lattice+k*spacedim;
    Ljptr=Lattice+j*spacedim;
    for(i=0;i<spacedim;i++) {
      if (LatNonZero(*Ljptr)) {
	LatMul(*LatTmp,*RoundedQuo,*Ljptr);
	LatSub(*Lptr,*Lptr,*LatTmp);
      }
      Lptr++; Ljptr++;
    }
    
    //  Update exact scalar products
    Lptr=ExactScals+k*allocdim+k;
    Ljptr=ExactScals+k*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      LatMul(*LatTmp,*RoundedQuo,*Ljptr);
      LatAdd(*LatTmp,*LatTmp,*LatTmp);
      LatSub(*Lptr,*Lptr,*LatTmp);
    }
    LatMul(*LatTmp,*RoundedQuo,*RoundedQuo);
    Ljptr=ExactScals+j*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      LatMul(*LatTmp,*LatTmp,*Ljptr);
      LatAdd(*Lptr,*Lptr,*LatTmp);
    }
    
    for(i=0;i<=maxk;i++) {
      if (i==k) continue;
      if (i<k) 
	Lptr=ExactScals+k*allocdim+i;
      else
	Lptr=ExactScals+i*allocdim+k;
      if (i<j) 
	Ljptr=ExactScals+j*allocdim+i;
      else
	Ljptr=ExactScals+i*allocdim+j;
      if (LatNonZero(*Ljptr)) {
	LatMul(*LatTmp,*RoundedQuo,*Ljptr);
	LatSub(*Lptr,*Lptr,*LatTmp);
      }
    }
  }
}

void FullUpdateOne(int j, int k, int maxk, LLL_PARAMS Params)
{
  int i;
  LatEntry *Lptr, *Ljptr;
  
  LatEntry *Lattice, *ExactScals, *LatTmp;
  GramEntry *GramM, *GramR, MinusOne, Tmp;
  int spacedim, allocdim;

  Lattice=Params.Lattice; ExactScals=Params.ExactScals; GramR=Params.GramR;
  GramM=Params.GramM; spacedim=Params.spacedim;
  allocdim=Params.allocdim;
  LatTmp=Params.LatTmp_ptr; MinusOne=Params.MinusOne; Tmp=Params.Tmp;
  
  for(i=0;i<j;i++) {
    GramCpy(Tmp,*(GramM+k*allocdim+i));
    GramSub(*(GramM+k*allocdim+i),Tmp,*(GramM+j*allocdim+i));
    GramCpy(Tmp,*(GramR+k*allocdim+i));
    GramSub(*(GramR+k*allocdim+i),Tmp,*(GramR+j*allocdim+i));
  }
  GramCpy(Tmp,*(GramM+k*allocdim+i));
  GramAdd(*(GramM+k*allocdim+i),Tmp,MinusOne);
  GramCpy(Tmp,*(GramR+k*allocdim+i));
  GramSub(*(GramR+k*allocdim+i),Tmp,*(GramR+j*allocdim+i));

  Lptr=Lattice+k*spacedim;
  Ljptr=Lattice+j*spacedim;
  for(i=0;i<spacedim;i++) {
    if (LatNonZero(*Ljptr)) {
      LatSub(*Lptr,*Lptr,*Ljptr);
    }
    Lptr++; Ljptr++;
  }
  
  //  Update exact scalar products
  Lptr=ExactScals+k*allocdim+k;
  Ljptr=ExactScals+k*allocdim+j;
  if (LatNonZero(*Ljptr)) {
    LatAdd(*LatTmp,*Ljptr,*Ljptr);
    LatSub(*Lptr,*Lptr,*LatTmp);
  }
  Ljptr=ExactScals+j*allocdim+j;
  if (LatNonZero(*Ljptr)) {
    LatAdd(*Lptr,*Lptr,*Ljptr);
  }
  
  for(i=0;i<=maxk;i++) {
    if (i==k) continue;
    if (i<k) 
      Lptr=ExactScals+k*allocdim+i;
    else
      Lptr=ExactScals+i*allocdim+k;
    if (i<j) 
      Ljptr=ExactScals+j*allocdim+i;
    else
      Ljptr=ExactScals+i*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      LatSub(*Lptr,*Lptr,*Ljptr);
    }
  }
}


void FullUpdateMinusOne(int j, int k, int maxk, LLL_PARAMS Params)
{ 
  int i;
  LatEntry *Lptr, *Ljptr;
  
  LatEntry *Lattice, *ExactScals, *LatTmp;
  GramEntry *GramM, *GramR, MinusOne, Tmp;
  int spacedim, allocdim;

  Lattice=Params.Lattice; ExactScals=Params.ExactScals; GramR=Params.GramR;
  GramM=Params.GramM; spacedim=Params.spacedim;
  allocdim=Params.allocdim;  
  LatTmp=Params.LatTmp_ptr; MinusOne=Params.MinusOne; Tmp=Params.Tmp;

  for(i=0;i<j;i++) {
    GramCpy(Tmp,*(GramM+k*allocdim+i));
    GramAdd(*(GramM+k*allocdim+i),Tmp,*(GramM+j*allocdim+i));
    GramCpy(Tmp,*(GramR+k*allocdim+i));
    GramAdd(*(GramR+k*allocdim+i),Tmp,*(GramR+j*allocdim+i));
  }
  GramCpy(Tmp,*(GramM+k*allocdim+i));
  GramSub(*(GramM+k*allocdim+i),Tmp,MinusOne);
  GramCpy(Tmp,*(GramR+k*allocdim+i));
  GramAdd(*(GramR+k*allocdim+i),Tmp,*(GramR+j*allocdim+i));
  
  Lptr=Lattice+k*spacedim;
  Ljptr=Lattice+j*spacedim;
  for(i=0;i<spacedim;i++) {
    if (LatNonZero(*Ljptr)) {
      LatAdd(*Lptr,*Lptr,*Ljptr);
    }
    Lptr++; Ljptr++;
  }
  
  //  Update exact scalar products
  Lptr=ExactScals+k*allocdim+k;
  Ljptr=ExactScals+k*allocdim+j;
  if (LatNonZero(*Ljptr)) {
    LatAdd(*LatTmp,*Ljptr,*Ljptr);
    LatAdd(*Lptr,*Lptr,*LatTmp);
  }
  Ljptr=ExactScals+j*allocdim+j;
  if (LatNonZero(*Ljptr)) {
    LatAdd(*Lptr,*Lptr,*Ljptr);
  }
  
  for(i=0;i<=maxk;i++) {
    if (i==k) continue;
    if (i<k) 
      Lptr=ExactScals+k*allocdim+i;
    else
      Lptr=ExactScals+i*allocdim+k;
    if (i<j) 
      Ljptr=ExactScals+j*allocdim+i;
    else
      Ljptr=ExactScals+i*allocdim+j;
    if (LatNonZero(*Ljptr)) {
      LatAdd(*Lptr,*Lptr,*Ljptr);
    }
  }
}

void InitExactScals(int k, int maxk, LLL_PARAMS Params)
{
  int i,j;
  LatEntry *Lattice, *ScalProd, *ExactScals;
  int spacedim, allocdim;
  LatEntry *LatTmp, *exScalar;
  LatEntry *Lptr, *Ljptr;

  ScalProd=Params.ScalProd;
  Lattice=Params.Lattice; ExactScals=Params.ExactScals; spacedim=Params.spacedim;
  LatTmp=Params.LatTmp_ptr; exScalar=Params.exScalar_ptr; allocdim=Params.allocdim;

  if (ScalProd==NULL) {
    for(j=0;j<=maxk;j++) {
      Lptr=Lattice+k*spacedim;
      Ljptr=Lattice+j*spacedim;
      ZeroLatEntry(*exScalar);
      for(i=0;i<spacedim;i++) {
	if ((LatNonZero(*Lptr)) && (LatNonZero(*Ljptr))) {
	  LatMul(*LatTmp,*Lptr,*Ljptr);
	  LatAdd(*exScalar,*exScalar,*LatTmp);
	}
	Lptr++; Ljptr++;
      }
      if (j<=k)
	LatSwap(*(ExactScals+k*allocdim+j),*exScalar);
      else
	LatSwap(*(ExactScals+j*allocdim+k),*exScalar);
    }
  }
  else {
    for(j=0;j<spacedim;j++) {
      Lptr=Lattice+k*spacedim;
      Ljptr=ScalProd+j*spacedim;
      ZeroLatEntry(*exScalar);
      for(i=0;i<spacedim;i++) {
	if ((LatNonZero(*Lptr)) && (LatNonZero(*Ljptr))) {
	  LatMul(*LatTmp,*Lptr,*Ljptr);
	  LatAdd(*exScalar,*exScalar,*LatTmp);
	}
	Lptr++; Ljptr++;
      }
      LatSwap(*(ScalProd+spacedim*spacedim+j),*exScalar);
    }
    
    for(j=0;j<=maxk;j++) {
      Lptr=ScalProd+spacedim*spacedim;
      Ljptr=Lattice+j*spacedim;
      ZeroLatEntry(*exScalar);
      for(i=0;i<spacedim;i++) {
	if ((LatNonZero(*Lptr)) && (LatNonZero(*Ljptr))) {
	  LatMul(*LatTmp,*Lptr,*Ljptr);
	  LatAdd(*exScalar,*exScalar,*LatTmp);
	}
	Lptr++; Ljptr++;
      }
      if (j<=k) {
	LatSwap(*(ExactScals+k*allocdim+j),*exScalar);
      }
      else {
	LatSwap(*(ExactScals+j*allocdim+k),*exScalar);
      }
    }
  }
}


void InitExactScalsNoise(int k, LLL_PARAMS Params)
{
  int i,j;
  LatEntry *Lattice, *ExactScalsNoise;
  int spacedim, allocdim;
  LatEntry *LatTmp, *exScalar;
  LatEntry *Lptr, *Ljptr;

  Lattice=Params.Lattice; ExactScalsNoise=Params.ExactScalsNoise; spacedim=Params.spacedim;
  LatTmp=Params.LatTmp_ptr; exScalar=Params.exScalar_ptr; allocdim=Params.allocdim;

  ZeroLatEntry(*LatTmp);
  Lptr=Lattice+k*spacedim;
  for(i=0;i<spacedim;i++) {
    if (((*Lptr)->_mp_size)<0) {
      LatSub(*LatTmp,*LatTmp,*Lptr);
    }
    else {
      LatAdd(*LatTmp,*LatTmp,*Lptr);
    }
    Lptr++;
  }

  for(j=0;j<=k;j++) {
    Ljptr=Lattice+j*spacedim;
    ZeroLatEntry(*exScalar);
    for(i=0;i<spacedim;i++) {
      if (((*Ljptr)->_mp_size)<0) {
	LatSub(*exScalar,*exScalar,*Ljptr);
      }
      else {
	LatAdd(*exScalar,*exScalar,*Ljptr);
      }
      Ljptr++;
    }
    LatMul(*(ExactScalsNoise+k*allocdim+j),*exScalar,*LatTmp);
  }
}


int DoLovaszExchanges(int *k_ptr, int maxk, int freshmupoint, LLL_PARAMS Params, int mupolicy)
{
  int save_k;
  int  cl_sg;
  int i, j, k, spacedim, allocdim;
  LatEntry *Lptr, *Ljptr, *Lattice, *ExactScals;
  GramEntry *GramM, *GramR, *GramS, LLLCste, Tmp, Scalneg, mu, mubis;
  int *muinvalidfrom;
  GramEntry swap_tmp;


  Lattice=Params.Lattice; ExactScals=Params.ExactScals; GramR=Params.GramR;
  GramM=Params.GramM; GramS=Params.GramS; LLLCste=Params.LLLCste; spacedim=Params.spacedim;
  allocdim=Params.allocdim; muinvalidfrom=Params.muinvalidfrom; mu=Params.mu; mubis=Params.mubis;
  Tmp=Params.Tmp;  Scalneg=Params.Scalneg;

  k=*k_ptr;
  save_k=k;
  for (;;) {
    GramMul(Scalneg,LLLCste,*(GramR+(k-1)*allocdim+(k-1)));
    GramSub(Tmp,*(GramS+k-1),Scalneg);

    cl_sg=ClearSign(Tmp);
    if (cl_sg>0){
      k++;
      (*k_ptr)=k;
      return 0;
    }
    
    if (cl_sg<0) {
      cl_sg=-cl_sg;
    }
    else {
#if 0
      GramMul(Scalneg,LLLCste,*(GramS+k-1));
      GramSub(Tmp,*(GramR+(k-1)*allocdim+(k-1)),Scalneg);
      cl_sg=ClearSign(Tmp);
#else
      // fprintf(stderr,"Would have performed degradation exchange !\n");
#endif
    }
    if (cl_sg>0) {
      //  Do Exchange !

      // Lattice elements
      Lptr=Lattice+k*spacedim;
      Ljptr=Lattice+(k-1)*spacedim;
      for(i=0;i<spacedim;i++) {
	LatSwap(*Lptr,*Ljptr);
	Lptr++; Ljptr++;
      }
      
      // ExactScals elements
      Lptr=ExactScals+k*allocdim;
      Ljptr=ExactScals+(k-1)*allocdim;
      for(i=0;i<k-1;i++) {
	LatSwap(*Lptr,*Ljptr);
	Lptr++; Ljptr++;
      }
      LatSwap(*(Lptr+1),*Ljptr);
      Lptr=ExactScals+(k+1)*allocdim+k;
      for(i=k+1;i<=maxk;i++) {
	LatSwap(*Lptr,*(Lptr-1));
	Lptr+=allocdim;
      }

      for(i=0;i<k-1;i++) {
	swap_tmp=*(GramM+(k-1)*allocdim+i);*(GramM+(k-1)*allocdim+i)=*(GramM+k*allocdim+i);*(GramM+k*allocdim+i)=swap_tmp;
	swap_tmp=*(GramR+(k-1)*allocdim+i);*(GramR+(k-1)*allocdim+i)=*(GramR+k*allocdim+i);*(GramR+k*allocdim+i)=swap_tmp;
      }
      GramCpy(Tmp,*(GramS+k-1));
      GramCpy(*(GramS+k-1),*(GramR+(k-1)*allocdim+(k-1)));
      GramCpy(*(GramR+(k-1)*allocdim+(k-1)),Tmp);
      
      GramCpy(mu,*(GramM+k*allocdim+k-1));
      GramDivide(*(GramM+k*allocdim+k-1),*(GramR+k*allocdim+k-1),*(GramR+(k-1)*allocdim+(k-1)));
      GramCpy(mubis,*(GramM+k*allocdim+k-1));
      GramMul(Tmp,*(GramM+k*allocdim+k-1),*(GramR+k*allocdim+k-1));
      GramSub(*(GramR+k*allocdim+k),*(GramS+k-1),Tmp);
      
// Don't bother to update mus -- just invalidate them
      for (i=k+1;i<=maxk;i++) {
	if (muinvalidfrom[i]>=k)
	  muinvalidfrom[i]=k-1;
      }
      
      k--;
      if (k==0) {
	*(k_ptr)=1;
	return 0;
      }
    }
    else {
      if (k<save_k) {
	// Make sure we recompute Lovasz for lower values
	k--;
	if (k<=0) k=1;
	*(k_ptr)=k;
	return 0;
      }
      else {
	if ((freshmupoint>=k)) {
	  // We need to increase precision here
	  if (VERBOSE) fprintf(stderr,"Lovasz condition not defined in LLL => increase prec\n");
	  *(k_ptr)=k;
	  return 1;
	}

	// if mu are not freshly computed, retry with recomputed values
	if (mupolicy==MU_LAZY_RECOMPUTE){
	  k=freshmupoint+1;
	  for(j=k;j<=maxk;j++) 
	    muinvalidfrom[j]=0;
	}
	*(k_ptr)=k;
	return 0;
      }
    }
  }
}


int DoDeepLovasz(int *k_ptr, int maxk, int freshmupoint, LLL_PARAMS Params, int mupolicy)
{
  int save_k;
  int  cl_sg;
  int i, j, k, spacedim, allocdim, depth, save_depth;
  LatEntry *Lptr, *Ljptr, *Lattice, *ExactScals;
  GramEntry *GramM, *GramR, *GramS, LLLCste, Tmp,  Scalneg, mu, mubis;
  int *muinvalidfrom;
  GramEntry swap_tmp;


  Lattice=Params.Lattice; ExactScals=Params.ExactScals; GramR=Params.GramR;
  GramM=Params.GramM; GramS=Params.GramS; LLLCste=Params.LLLCste; spacedim=Params.spacedim;
  allocdim=Params.allocdim; muinvalidfrom=Params.muinvalidfrom; mu=Params.mu; mubis=Params.mubis;
  Tmp=Params.Tmp; Scalneg=Params.Scalneg; 
  
  k=*k_ptr;
  save_k=k;
  for (;;) {
    // Use mu and mubis to store products
    save_depth=k+1;
    for(depth=1;depth<=k;depth++) {
      if (depth==1) {
	GramCpy(mu,*(GramR+(k-1)*allocdim+(k-1)));
	GramCpy(mubis,*(GramS+k-1)); 
     }
      else {
	GramCpy(Tmp,mu);
	GramMul(mu,Tmp,*(GramR+(k-depth)*allocdim+(k-depth)));
	GramCpy(Tmp,mubis);
	GramMul(mubis,Tmp,*(GramS+k-depth));
      }
      GramMul(Scalneg,LLLCste,mu);
      GramSub(Tmp,mubis,Scalneg);
      cl_sg=ClearSign(Tmp);
      if (cl_sg<0) {
	save_depth=depth;
	break; // Experimentally, it seems better to break here !
      }
    }
    depth=save_depth;
    if (depth>k) {
      // Check normal Lovasz condition
      GramMul(Scalneg,LLLCste,*(GramR+(k-1)*allocdim+(k-1)));
      GramSub(Tmp,*(GramS+k-1),Scalneg);
      cl_sg=ClearSign(Tmp);
      if (cl_sg>0){
	k++;
	(*k_ptr)=k;
	return 0;
      }
#if 0      
      GramMul(Scalneg,LLLCste,*(GramS+k-1));
      GramSub(Tmp,*(GramR+(k-1)*allocdim+(k-1)),Scalneg);
      cl_sg=ClearSign(Tmp);
      if (cl_sg>0) depth=1;
#endif
      else {
	if (k<save_k) {
	  // Recompute Lovasz for lower values
	  k--;
	  if (k<=0) k=1;
	  *(k_ptr)=k;
	  return 0;
	}
	else {
	  if ((freshmupoint>=k)) {
	    fprintf(stderr,"Lovasz condition not defined in LLL\n");
	    *(k_ptr)=k;
	    return 1;
	  }

	  if (mupolicy==MU_LAZY_RECOMPUTE){
	    k=freshmupoint+1;
	    for(j=k;j<=maxk;j++) 
	      muinvalidfrom[j]=0;
	  }
	  *(k_ptr)=k;
	  return 0;
	}
      }
    }

    for(;depth>0;depth--) {
      // Do the exchange(s)
      // Lattice elements
      Lptr=Lattice+k*spacedim;
      Ljptr=Lattice+(k-1)*spacedim;
      for(i=0;i<spacedim;i++) {
	LatSwap(*Lptr,*Ljptr);
	Lptr++; Ljptr++;
      }
      
      // ExactScals elements
      Lptr=ExactScals+k*allocdim;
      Ljptr=ExactScals+(k-1)*allocdim;
      for(i=0;i<k-1;i++) {
	LatSwap(*Lptr,*Ljptr);
	Lptr++; Ljptr++;
      }
      LatSwap(*(Lptr+1),*Ljptr);
      Lptr=ExactScals+(k+1)*allocdim+k;
      for(i=k+1;i<=maxk;i++) {
	LatSwap(*Lptr,*(Lptr-1));
	Lptr+=allocdim;
      }
      
      for(i=0;i<k-1;i++) {
	swap_tmp=*(GramM+(k-1)*allocdim+i);*(GramM+(k-1)*allocdim+i)=*(GramM+k*allocdim+i);*(GramM+k*allocdim+i)=swap_tmp;
	swap_tmp=*(GramR+(k-1)*allocdim+i);*(GramR+(k-1)*allocdim+i)=*(GramR+k*allocdim+i);*(GramR+k*allocdim+i)=swap_tmp;
      }
      GramCpy(Tmp,*(GramS+k-1));
      GramCpy(*(GramS+k-1),*(GramR+(k-1)*allocdim+(k-1)));
      GramCpy(*(GramR+(k-1)*allocdim+(k-1)),Tmp);
      
      GramCpy(mu,*(GramM+k*allocdim+k-1));
      GramDivide(*(GramM+k*allocdim+k-1),*(GramR+k*allocdim+k-1),*(GramR+(k-1)*allocdim+(k-1)));
      GramCpy(mubis,*(GramM+k*allocdim+k-1));
      GramMul(Tmp,*(GramM+k*allocdim+k-1),*(GramR+k*allocdim+k-1));
      GramSub(*(GramR+k*allocdim+k),*(GramS+k-1),Tmp);
      
      // Don't bother to update mus
      for (i=k;i<=maxk;i++) {
	if (muinvalidfrom[i]>=k)
	  muinvalidfrom[i]=k-1;
      }
      k--;
      if (k==0) {
	*(k_ptr)=1;
	return 0;
      }
    }
  }
}


int ComputeMuValues(int k, LLL_PARAMS Params)
{
  int j;
  int allocdim;
  LatEntry *ExactScals,*ExactScalsNoise;
  GramEntry *GramM, *GramR, Tmp, Scal, Scalpos;
  int *muinvalidfrom;


  ExactScals=Params.ExactScals; GramR=Params.GramR;
  ExactScalsNoise=Params.ExactScalsNoise;
  GramM=Params.GramM; 
  allocdim=Params.allocdim; muinvalidfrom=Params.muinvalidfrom; Scal=Params.Scal;
  Tmp=Params.Tmp; Scalpos=Params.Scalpos;

  for(j=muinvalidfrom[k];j<k;j++) {
      LowLevelExactMinusDotProduct(Scal,*(ExactScals+k*allocdim+j),GramM+k*allocdim, GramR+j*allocdim, j);
    if (ExactScalsNoise!=NULL) {
      AddLatEntryToGramNoise(Scal,*(ExactScalsNoise+k*allocdim+j),Tmp,Scalpos);
    }
  if (GramDivide(*(GramM+k*allocdim+j),Scal,*(GramR+j*allocdim+j))) {
    fprintf(stderr,"Division by zero\n");
    return 1;
  }
  GramCpy(*(GramR+k*allocdim+j),Scal);
  }
  muinvalidfrom[k]=k;

  return 0;
}

void ComputeNormValues(int k, LLL_PARAMS Params)
{
  int j;
  int allocdim;
  LatEntry *ExactScals,*ExactScalsNoise;
  GramEntry *GramM, *GramR, *GramS, Tmp, Scal;


  ExactScals=Params.ExactScals; GramR=Params.GramR;
  ExactScalsNoise=Params.ExactScalsNoise;
  GramM=Params.GramM; GramS=Params.GramS; 
  Scal=Params.Scal; Tmp=Params.Tmp;
  allocdim=Params.allocdim;


  LatEntryToGram(*(GramS),*(ExactScals+k*allocdim+k));
  if (ExactScalsNoise!=NULL) {
    AddLatEntryToGramNoise(*(GramS),*(ExactScalsNoise+k*allocdim+k),Tmp,Scal);
  }
  for(j=0;j<k;j++) {
    GramMul(Tmp,*(GramM+k*allocdim+j),*(GramR+k*allocdim+j));
    GramSub(*(GramS+j+1),*(GramS+j),Tmp);
  }
  GramCpy(*(GramR+k*allocdim+k),*(GramS+k));
}


int DoSizeReduction(int k, int maxk, int discoverystep, int *freshmupoint, int mupolicy, LLL_PARAMS Params)
{ 
  int nonzero, j, status, glob_status;
  int allocdim;
  LatEntry *RoundedQuo;
  GramEntry *GramM, Quo;
  int *muinvalidfrom;


  GramM=Params.GramM; RoundedQuo=Params.RoundedQuo_ptr; 
  allocdim=Params.allocdim; muinvalidfrom=Params.muinvalidfrom;
  Quo=Params.Quo;

  glob_status=0;
  nonzero=0;
    
  // Size reduction (approx and iterate)
  for(j=k-1;j>=0;j--) {
    status=DoCompareNoisyHalf(*(GramM+k*allocdim+j));

    if (status==1) {
      if ((*freshmupoint)==k) {
	(*freshmupoint)--;
	muinvalidfrom[k]=0;
      }
      if (mupolicy==MU_LAZY_RECOMPUTE){
	if (discoverystep) muinvalidfrom[k]=0;
	else if (muinvalidfrom[k]>j) muinvalidfrom[k]=j;
      }
      
      RoundGramEntryToLat(*RoundedQuo, *(GramM+k*allocdim+j));
      LatEntryToGram(Quo, *RoundedQuo);
      nonzero=1;
	
      if (!LatNonZero(*RoundedQuo)) {
	fprintf(stderr,"Should not be possible -- Please report !\n");
	exit(1);
      }
      
      if (LatCmpCste(*RoundedQuo,1)==0) {
	FullUpdateOne(j, k, maxk, Params);
      }
      else if (LatCmpCste(*RoundedQuo,-1)==0) {
	FullUpdateMinusOne(j, k, maxk, Params);
      }
      else {
	FullUpdate(j,k, maxk, Params);
      }
    }
    else if (status==0) glob_status=1;
  }

  if (nonzero) return 1;
  if (glob_status) return 2;
  return 0;
}


int LL_allocDone(LLL_PARAMS Params, int *finaldim,
		 int *discoverymax,  int mupolicy)		 
{
  LatEntry *Lattice, *ExactScals, *ExactScalsNoise;
  GramEntry *GramR, *GramM;
  int spacedim, allocdim, dim;
  int *muinvalidfrom;
  
  int k, save_k, freshmupoint;
  int i, j, maxk, mink, prevk, glob_status;
  int discoverystep;
  int totiters;
  LatEntry *Lptr, *Ljptr;

  Lattice=Params.Lattice; ExactScals=Params.ExactScals; GramR=Params.GramR;
  ExactScalsNoise=Params.ExactScalsNoise;
  GramM=Params.GramM; 
  spacedim=Params.spacedim;
  allocdim=Params.allocdim; dim=Params.dim; muinvalidfrom=Params.muinvalidfrom;
  
  k             = 0;   maxk          = -1;    mink          = 0;
  prevk         = -1;  totiters      = 0;
  *finaldim     = dim; freshmupoint  = 0;     discoverystep = 0;

  while (k <= dim) {
    if (userwantexit>=2) break;
    if (prevk != k) {
      prevk = k;
      discoverystep = 0;
    }
    totiters++;
    
    if (k<mink) mink=k;
    if (k<=1) freshmupoint=0;
    if (k>maxk) {
      if ((k>freshmupoint+1)&&(mupolicy==MU_LAZY_RECOMPUTE)){
	k=freshmupoint+1;
	for(j=k;j<=maxk;j++) muinvalidfrom[j]=0;
	continue;
      }
      if (k==dim) break;
      if (k>*discoverymax) {
	discoverystep=1; 
	*discoverymax=k-1; if ((*discoverymax)<0) *discoverymax=0;
      }
      maxk=k;
      if (VERBOSE) {
	if (k==0) 
	  fprintf(stderr,"maxk=%d (mink=%d dim=%d) [%f / %d]\n",
		  maxk,mink,dim,clock()*1.0/CLOCKS_PER_SEC, totiters);
	else 
	  gmp_fprintf(stderr,"maxk=%d (freshmu=%d) (mink=%d dim=%d) [%f / %d] size=%Zd\n",
		      maxk,freshmupoint,mink,dim,clock()*1.0/CLOCKS_PER_SEC, totiters,*ExactScals);
	fflush(stderr);
      }
      mink=maxk;
      
      /* Initialize exact scalar products at maxk */
      InitExactScals(maxk, maxk, Params);
    }
    
    /* At first vector, copy exact data into Gram, increment and move */
    if (k==0) { 
      LatEntryToGram(*GramR, *ExactScals);
      k++;
      continue;
    }
    
    /* Need to test zero vector to deal with generating families*/
    if (LatNonZero(*(ExactScals+k*allocdim+k))<0) {
      fprintf(stderr,"Found vector with negative norm, bad scalar product !\n");
      return 0;
    }
    
    if (!LatNonZero(*(ExactScals+k*allocdim+k))) {
      if (VERBOSE) {
	fprintf(stderr,"Zero vector in Lattice : Tweak maxk for simplicity\n");
      }
      for(i=k;i<dim-1;i++) {
	Lptr=Lattice+i*spacedim;
	Ljptr=Lattice+(i+1)*spacedim;
	for(j=0;j<spacedim;j++) {
	  LatSwap(*Lptr,*Ljptr);
	  Lptr++; Ljptr++;
	}
      }
      dim--;
      *finaldim=dim;
      maxk=k-1;
      continue;
    }
    
    if (mupolicy==MU_MAX_RECOMPUTE) muinvalidfrom[k]=0;
    if ((freshmupoint==k-1)&&(muinvalidfrom[k]==0)) {
      freshmupoint=k;
    }
    
    if (ExactScalsNoise!=NULL) {
      // Fully recompute ExactScalsNoise for k
      InitExactScalsNoise(k, Params);
    }
    
    glob_status=ComputeMuValues(k, Params);
    
    if (glob_status==1) {
      if (freshmupoint==k) return 1;
      for(j=0;j<=maxk;j++) muinvalidfrom[j]=0;
      freshmupoint=0;
      k=1;
      continue;
    }
    
    glob_status=DoSizeReduction(k, maxk, discoverystep, &freshmupoint, mupolicy, Params);
    
    if (glob_status==1) {
      continue;
    }
    else if (glob_status==2) {
      if (freshmupoint>=k) {
	// Insufficient precision detected here with fresh Mu's
	// Skip and detect problem in Lovasz condition
      }
      else {
	k=freshmupoint+1;
	if (mupolicy==MU_LAZY_RECOMPUTE){
	  for(j=k;j<=maxk;j++) 
	    muinvalidfrom[j]=0;
	}
	continue;
      }
    }
    
    ComputeNormValues(k, Params);
    
    
#define EARLY_PREC_THRESHOLD 20
    if (EARLY_PREC_DETECT && (discoverystep==0)) {
      // Heuristic to pre detect precision loss
      // (Faster than waiting until precision if definitely insufficient.)
      // Please report if this causes any failure
      // Can be desactivated on the command line by specifying -n
      if (((*(GramM+k*allocdim+k-1))->sign)&&(my_mpn_bitnum((*(GramM+k*allocdim+k-1))->mantissa,flti_glob_nblimbs)<EARLY_PREC_THRESHOLD)) {
	fprintf(stderr,"Precision seems to get to low ... \n");
	return 1;
      }
    }
    
    save_k=k;
    if (DEEP_LOVASZ) {
      if (DoDeepLovasz(&k, maxk, freshmupoint, Params, mupolicy)) return 1;
    }
    else {
      if (DoLovaszExchanges(&k, maxk, freshmupoint, Params, mupolicy)) return 1;
    }
    
    if (k<=save_k) {
      if (freshmupoint>=k) freshmupoint=k-1;
      if (freshmupoint<0) freshmupoint=0;
    }
  }

  if (VERBOSE) {
    gmp_fprintf(stderr,"Final (freshmu=%d) (dim=%d) [%f / %d] size=%Zd\n",
		freshmupoint,dim,clock()*1.0/CLOCKS_PER_SEC, totiters,*ExactScals);
  }
  
  return 0;
}

int LL_fixedPrec(LatEntry *Lattice, LatEntry *ScalProd,  LatEntry *ExactScals, LatEntry *ExactScalsNoise,
		 GramEntry *GramR, GramEntry *GramM, GramEntry *GramS,
		 int spacedim, int allocdim, int dim, int *finaldim, double csteLLL,
		 int *discoverymax, int mupolicy)
 {
  int *muinvalidfrom,retcode;

  GramEntry Scal, Scalneg, Scalpos, MinusOne, Tmp, Quo, LLLCste, mu, mubis, muter;
  LatEntry RoundedQuo, LatTmp, exScalar;

  LLL_PARAMS Params;
  
  muinvalidfrom=(int *) calloc(allocdim,sizeof(int));
  if (muinvalidfrom==NULL) {
    fprintf(stderr,"malloc failed!\n");
    return 1;
  }

  extern flti flti_allocate(long count);

  /* Initialization part */
  mu=flti_allocate(1);
  Tmp=flti_allocate(1);
  Quo=flti_allocate(1);
  Scal=flti_allocate(1);
  mubis=flti_allocate(1);
  muter=flti_allocate(1);
  LLLCste=flti_allocate(1);
  Scalneg=flti_allocate(1);
  Scalpos=flti_allocate(1);
  MinusOne=flti_allocate(1);
  if ((mu==NULL) || (Tmp==NULL) || (Quo==NULL) || (Scal==NULL) ||
      (mubis==NULL) || (muter==NULL) || (LLLCste==NULL) || (Scalneg==NULL) ||
      (Scalpos==NULL) || (MinusOne==NULL)) {
    fprintf(stderr,"Allocation failed !\n");
  }
  
  InitLatEntry(LatTmp);
  InitLatEntry(exScalar);
  InitLatEntry(RoundedQuo);

  // Set Minus One to Exact constant ! 
  ZeroGramEntry(MinusOne);
  MinusOne->sign=-1;
  (MinusOne->mantissa)[0]=1;
  flti_normalize(MinusOne);

  // Set LLL Cst to interval arithmeric and erase error ! 
  GramSetFromCst(LLLCste,csteLLL);
  LLLCste->errormant=0;
  flti_normalize(LLLCste);


  Params.Lattice=Lattice; Params.ScalProd=ScalProd; Params.ExactScals=ExactScals;
  Params.ExactScalsNoise=ExactScalsNoise;
  Params.GramR=GramR; Params.GramM=GramM; Params.GramS=GramS; 
  Params.spacedim=spacedim; Params.allocdim=allocdim; Params.dim=dim; 
  Params.mu=mu; Params.Tmp=Tmp; Params.Quo=Quo; 
  Params.Scal=Scal; Params.mubis=mubis; Params.muter=muter; 
  Params.LLLCste=LLLCste; Params.Scalneg=Scalneg; Params.Scalpos=Scalpos; 
  Params.MinusOne=MinusOne; Params.LatTmp_ptr=&LatTmp; Params.exScalar_ptr=&exScalar;
  Params.RoundedQuo_ptr=&RoundedQuo; Params.muinvalidfrom=muinvalidfrom;


  retcode=LL_allocDone(Params, finaldim, discoverymax, mupolicy);
  
  /* Cleaning up */

  free(Scal);
  free(Scalneg);
  free(Scalpos);
  free(MinusOne);
  free(Tmp);
  free(Quo);
  free(LLLCste);
  free(mu);
  free(mubis);
  free(muter);

  DelLatEntry(RoundedQuo);
  DelLatEntry(LatTmp);
  DelLatEntry(exScalar);
  free(muinvalidfrom);

  return retcode;
}

void handler(int ignorevar)
{
  if (userwantexit==0) {
    userwantexit=1;
    fprintf(stderr,"Will cleanly quit after next full LLL ...\n");
  }
  else if (userwantexit==1) {
    userwantexit=2;
    fprintf(stderr,"Will cleanly quit after next LLL step ...\n");
  }
  else {
    fprintf(stderr,"Immediately aborting\n");
    exit(1);
  }
}


int LL_Adaptative(LatEntry *Lattice, LatEntry *ScalProd, int noisy,
		  int spacedim, int dim, int allocdim,
		  int *finaldim, double csteLLL, int mupolicy)
{

  int errcode,prec,precinc, maxk;
  LatEntry *ExactScals,*ExactScalsNoise;
  GramEntry *GramR,*GramM,*GramS;

  struct sigaction action;

  sigemptyset(&action.sa_mask);
  action.sa_flags=0;
  action.sa_handler=&handler;
  sigaction(SIGINT,&action,NULL);
  sigaction(SIGTERM,&action,NULL);

  maxk=0;
  errcode=AllocLat(&ExactScals, allocdim, allocdim);
  if (errcode) {
    fprintf(stderr,"Error during ExactScals allocation\n");
    return 1;
  }
  if (noisy) {
    errcode=AllocLat(&ExactScalsNoise, allocdim, allocdim);
    if (errcode) {
      fprintf(stderr,"Error during ExactScals allocation\n");
      return 1;
    }
  }
  else ExactScalsNoise=NULL;
  
  for(prec=128,precinc=64;prec<=2048;prec+=precinc) {
    if (VERBOSE) fprintf(stderr,"precision=%d\n",prec);
    GramSetPrec(prec);
    
    errcode=AllocGram(&GramR, allocdim, allocdim);
    if (errcode) {
      fprintf(stderr,"Error during GramR allocation\n");
      return 1;
    }
    errcode=AllocGram(&GramM, allocdim, allocdim);
    if (errcode) {
      fprintf(stderr,"Error during GramM allocation\n");
      return 1;
    }

    errcode=AllocGram(&GramS, 2*allocdim, 1);
    if (errcode) {
      fprintf(stderr,"Error during GramS allocation\n");
      return 1;
    }

    errcode=LL_fixedPrec(Lattice, ScalProd, ExactScals, ExactScalsNoise,
			 GramR, GramM, GramS, 
			 spacedim, allocdim, dim, finaldim,
			 csteLLL, &maxk, mupolicy);
    
    UnAllocGram(&GramR, allocdim, allocdim);
    UnAllocGram(&GramM, allocdim, allocdim);
    UnAllocGram(&GramS, 2*allocdim, 1);
    if (errcode==0) break;
  }

  if (errcode) {
    fprintf(stderr,"Max precision insufficient (please report or adjust code)\n");
  }
  
  UnAllocLat(&ExactScals, allocdim, allocdim);
  if (noisy) {
    UnAllocLat(&ExactScalsNoise, allocdim, allocdim);
  }
  
  return errcode;
}

