#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include "LLL.h"

int main(int argc, char **argv)
{
  /*  Lattice Entries */
  LatEntry *Lattice;
  LatEntry *ScalProd;
  int dim, allocdim, spacedim, scaldim1, scaldim2, scaldim3;
  int i, enddim;
  int errcode;
  double lll_cste=0.99;
  int mupolicy=MU_MAX_RECOMPUTE;
  int withscalarprod=0;
  int noisy=0;


  if (argc>1) {
    for (i=1;i<argc;i++) {
      // read command line
      if (argv[i][0]!='-') {
	fprintf(stderr,"Unrecognized command line option %s\n",argv[i]);
	exit(1);
      }
      
      // -v...  = -verbose => VERBOSE=1
      if (argv[i][1]=='v') {
	VERBOSE=1;
	fprintf(stderr,"Verbose mode\n");
      }
      
      // -S...  = -Scalar prod => withscalarprod=1
      else if (argv[i][1]=='S') {
	withscalarprod=1;
	if (argv[i][2]=='n') {
	  noisy=1;
	}
      }
      // -p...  = -potential => DEEP_LOVASZ=1
      else if (argv[i][1]=='p') {
	DEEP_LOVASZ=1;
	fprintf(stderr,"Potential-LLL variant\n");
      }
      // -e...  = -early_prec_detect => EARLY_PREC_DETECT=1
      else if (argv[i][1]=='e') {
	EARLY_PREC_DETECT=1;
	fprintf(stderr,"Early precision defect detection activated\n");
      }
      //-d..  0.xx => set delta 
      else if (argv[i][1]=='d') {
	if ((i+1)<argc) {
	  i++;
	  lll_cste=0.0;
	  sscanf(argv[i],"%lf",&lll_cste);
	  if ((lll_cste<=0.25) || (lll_cste>=1.0)) {
	    fprintf(stderr,"Invalid command line option %s %s => lll_cste=%f\n",argv[i-1],argv[i],lll_cste);
	    exit(1);
	  }
	  else {
	    fprintf(stderr,"Setting lll_cste=%f\n",lll_cste);
	  }
	}
	else {
	  fprintf(stderr,"Option %s expects an argument\n",argv[i]);
	  exit(1);
	}
      }
      //-m... =-mupolicy   max/lazy => set mu_policy 
      else if (argv[i][1]=='m') {
	if ((i+1)<argc) {
	  i++;
	  if (argv[i][0]=='l') {
	    mupolicy=MU_LAZY_RECOMPUTE;
	  }
	  else if (argv[i][0]=='m') {
	    mupolicy=MU_MAX_RECOMPUTE;
	  }
	  else {
	    fprintf(stderr,"Argument %s is not a valid mu-policy\n",argv[i]);
	    exit(1);
	  }
	}
	else {
	  fprintf(stderr,"Option %s expects an argument\n",argv[i]);
	  exit(1);
	}
      }
      else {
	fprintf(stderr,"Unknown Option %s\n",argv[i]);
	exit(1);
      }
    }
  }

  errcode=AllocAndRead(&Lattice, &spacedim, &dim, &allocdim, 1, 1);
  if (errcode) {
    fprintf(stderr,"Error during lattice input\n");
    return 1;
  }

  if (withscalarprod) {
    if (noisy==0)
      fprintf(stderr,"Reading Scalar product matrix now\n");
    else 
      fprintf(stderr,"Reading Scalar product matrix now [Noisy mode]\n");
    errcode=AllocAndRead(&ScalProd, &scaldim1, &scaldim2, &scaldim3, 1, 1);
    if (errcode) {
      fprintf(stderr,"Error during scalar product input\n");
      return 1;
    }
    if ((scaldim1!=spacedim) || (scaldim2!=spacedim)) {
      fprintf(stderr,"Scalar product should be a square mat of dim %d\n",spacedim);
      return 1;
    }
  }
  else {
    ScalProd=NULL;
  }

  errcode=LL_Adaptative(Lattice, ScalProd, noisy, spacedim, dim, allocdim,
			&enddim, lll_cste, mupolicy);
  if (errcode) {
    fprintf(stderr,"LLL reports error\n");
    // return 1;
  }

  NTLWriteLattice(Lattice, spacedim, enddim);
  if (ScalProd!=NULL) NTLWriteLattice(ScalProd, spacedim, spacedim);
  return 0;
}
